import pickle

import numpy as np
from scipy.spatial.transform import Rotation


def simCamTransform(motion: str, X: np.ndarray, rotation_threshold: float = 0.8):
    """

    generation the rotation matrix and translation vector to apply a certain transformation (motion) to a point cloud

    :param motion: only "rotation" is implemented
    :param X: (6, npps)
    :return:
            rotation matrix, np.ndarray, (3,3)
            translation vector, np.ndarray, (3,1)
    """
    if motion == "rotation":
        angles = rotation_threshold * (
            0.5 - np.random.rand(1, 3)
        )  # it was 0.8, it has been 1.5 to train the single model, but I now parametrized (I'll start by using 0.8 again)
        R2 = Rotation.from_euler("zyx", angles, degrees=False)
        R2 = R2.as_matrix()[0]
        T2 = np.zeros((3, 1))

        return R2, T2


def simCamProj(camproj: str, X: np.ndarray, R: np.ndarray, t: np.ndarray):
    """
    generates a pair of 2D camera projections from a 3D point cloud.
    the first one is simply the projection on the z=1 plane of the point cloud.
    the second one is obtained by applying the rotation and translation given as inputs.

    :param camproj: kind of projection. only "uncalibrated" is implemented
    :param X: 3D point cloud, np.ndarray, (3, npps) (it is written in non-homogenenous 3D coordinates)
    :param R: rotation matrix, np.ndarray, (3,3)
    :param t: translation vector, np.ndarray, (3,1)
    :return:
        x1, 1st projection, np.ndarray, (2, npps)
        x2, 2nd projection, np.ndarray, (2, npps)
        K, intrinsic matrix np.ndarray (3,3)
    """

    if camproj == "uncalibrated":
        f = 0.25 + 5 * np.random.rand(1, 1)

        P1 = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0]])

        P2 = np.concatenate((np.transpose(R), np.matmul(-np.transpose(R), t)), axis=1)

        N = X.shape[1]

        homo_X = np.concatenate((X, np.ones((1, N))), axis=0)

        lx1 = np.matmul(P1, homo_X)  # project points in 1st camera
        lx2 = np.matmul(P2, homo_X)  # project points in 2nd camera
        x1 = f * lx1[0:2, :] / np.repmat(lx1[2, :], 2, 1)
        x2 = f * lx2[0:2, :] / np.repmat(lx2[2, :], 2, 1)

        K = np.eye(3)
        K[0, 0] = f
        K[1, 1] = f  # k does only a 2D scaling ( no 2D translation and 2D shear)

        return x1, x2, K


def normalize(x: np.ndarray, alpha: float):
    """

    :param x: 2d point cloud, np.ndarray, (2, npts)
    :param alpha: float, don't know actually
    :return:
        x_out, normalized 2D point cloud, np.ndarray, (2,nps)
        TN, de-normalization matrix?
    """
    nps = x.shape[1]  # num points
    # zero-center
    t = np.mean(x, axis=1)
    x_out = x - np.repeat(t.reshape((-1, 1)), nps, axis=1)

    # don't know actually
    x_out[0, :] = x_out[0, :] / alpha

    # scaling in such a way that the RMS distance from the origin is sqrt(2) --> the "average point" is (1,1)
    d1 = np.mean(np.sqrt(np.sum(x_out**2, axis=0)))
    s = np.sqrt(2) / d1
    x_out = x_out * s

    # compute T (transformation matrix to de-normalize x_out)
    A = np.array([[s, 0, 0], [0, s, 0], [0, 0, 1]])

    B = np.array([[1 / alpha, 0, -t[0] / alpha], [0, 1, -t[1]], [0, 0, 1]])

    TN = np.matmul(A, B)
    return x_out, TN


def residual(x1, x2, Hn):
    """
    check the residual of the prediction for the homography Hn
    :param x1:
    :param x2:
    :param Hn:
    :return:
    """
    N = x1.shape[1]
    x2_pred = np.matmul(Hn, x1)
    error = 0
    for point in range(N):
        distance = np.sqrt(np.sum((x2[:, point] - x2_pred[:, point]) ** 2))
        error = error + distance
    error /= N
    return error


def save_obj(obj, name):
    """
    loads into obj the pkl file identified by "name"
    :param obj: obj into which to load file
    :param name: path to file
    :return:
    """
    with open(name + ".pkl", "wb") as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_obj(name):
    """

    :param name: path to pkl file
    :return: loaded object
    """
    with open(name + ".pkl", "rb") as f:
        return pickle.load(f)



from typing import Tuple, Union

import numpy as np


def conic_monomials(points: np.ndarray) -> np.ndarray:
    """

    For each row (a point in homogeneous coordinates) in input array, compute the conic monomials.

    Input:
    [[x1, y1, 1],
     [x2, y2, 1]]

    Output:
    [[x1^2, x1y1, y1^2, x1, x2, 1],
      x2^2, x2y2, y2^2, x2, x1, 1]]

    :param points: np.ndarray, (num_points, 3).
    :return: np.ndarray, (num_points, 6)
    """
    n_points = points.shape[0]
    rows = np.zeros(shape=(n_points, 6))
    for i in range(n_points):

        x = points[i][0]
        y = points[i][1]
        rows[i, 0] = x**2
        rows[i, 1] = x * y
        rows[i, 2] = y**2
        rows[i, 3] = x
        rows[i, 4] = y
        rows[i, 5] = 1

    return rows


def circle_monomials(points: np.ndarray):
    """
    given a set of points returns a matrix whose rows are the conic extensions for each point
    specifically, the terms are:
    x^2 + y^2; x; y; 1

    :param points: np.ndarray, (num_points,). points are represented in homogeneous coordinates (x,y,z=1)
    :return: np.ndarray, (num_points, 4)
    """
    n_points = points.shape[0]
    rows = np.zeros(shape=(n_points, 4))
    for i in range(len(points)):
        x = points[i][0]
        y = points[i][1]
        row = np.zeros(shape=(4,))
        row[0] = x**2 + y**2
        row[1] = x
        row[2] = y
        row[3] = 1
        rows[i, :] = row

    return rows


def dlt_coefs(
    vandermonde: np.ndarray, weights: Union[np.ndarray, None] = None
) -> np.ndarray:
    """
    compute coefficients of a conic through Direct Linear Transformation.


    :param vandermonde: (number of points, number of monomials),np.ndarray or tf.tensor. each row contains monomials (e.g. for a conic: x^2 xy y^2 x y 1) of the corresponding point
    :param weights: (number of points,) np.ndarray or tf.tensor. probability of belonging to the model for each row in the vandermonde matrix.
                    if all points belong to the model don't specify its value.
    :return: np.ndarray, (number of monomials,), the coefficients computed via dlt
    """

    weights = weights + np.random.normal(0, 1e-9)
    weights = weights / np.linalg.norm(weights)
    weights = np.diag(weights)
    weighted_vander = np.matmul(weights, vandermonde)
    U, S, VT = np.linalg.svd(weighted_vander)
    V = np.transpose(VT)

    dlt_coefficients = V[:, -1]
    dlt_coefficients = dlt_coefficients * (
        1.0 / dlt_coefficients[0]
    )  # want the x^2 and y^2 terms to be close to 1
    return dlt_coefficients


def circle_coefs(radius: float, center: Tuple[float, float], verbose: bool = True):
    """
    given a radius and a center it returns the parameters of the conic

    :param radius: radius of the circle
    :param center: center (x,y) of the circle
    :param verbose: True: returns 6 coefficients, corresponding to the terms (x^2; xy; x^2; x; y; 1)
                    False: returns 4 coefficients, corresponding to the terms (x^2+y^2; x; y; 1)
    :return: np.ndarray, (num coefs,)
    """
    a = 1
    b = 0
    c = 1
    d = -2 * center[0]
    e = -2 * center[1]
    f = center[0] ** 2 + center[1] ** 2 - radius**2

    if verbose:
        return np.array([a, b, c, d, e, f], dtype=float)
    else:
        return np.array([a, d, e, f], dtype=float)


def veronese_map(points, n):
    """
    given a set of points and a degree n returns the veronese map of degree n for such points
    example:
    consider a veronese map of degree n for each homogeneous points (x,y,1) we'll have a row
    x^2 y^0 z^0 + x^1 y^1 z^0 + x^1 y^0 z^1 + x^0 y^2 z^0 + x^0 y^1 z^1 + x^0 y^0 z^2
    and, since z = 1, we can rewrite it as:
    x^2 + xy + x + y^2 + y + 1
    in tabular form:
      x    y    z
     i=2; j=0; k=0
     i=1; j=1; k=0
     i=1; j=0; k=1
     i=0; j=2; k=0
     i=0; j=1; k=1
     i=0; j=0; k=2


    :param points: np.ndarray, (num_points,)
    :param n: degree of veronese map
    :return: np.ndarray with veronese map, (num_points, veronese_columns)
    """
    cols = []
    i = n
    while i >= 0:
        j = n - i
        while j >= 0:
            k = n - i - j
            cols.append(points[:, 0] ** i * points[:, 1] ** j * points[:, 2] ** k)
            j -= 1
        i -= 1
    return np.transpose(np.array(cols))



"""A sample CLI."""

import click
import log


@click.command()
@click.argument("feet")
def main(text: str):
    log.init()
    click.echo(text)


if __name__ == "__main__":  # pragma: no cover
    main()  # pylint: disable=no-value-for-parameter



import math
import random
from typing import List, Tuple, Union

import numpy as np
from scipy.spatial.transform.rotation import Rotation
from sydraw.utils.cameraops import normalize, simCamProj, simCamTransform
from sydraw.utils.config import OPTS
from sydraw.utils.utils import compute_num_inliers_per_model


def outliers(
    x_range: Tuple[float, float],
    y_range: Tuple[float, float],
    n: int,
    homogeneous: bool = True,
) -> np.ndarray:
    """
    generates outliers in a user specified 2D square

    :param x_range: floats tuple, interval of x values
    :param y_range: floats tuple, interval if y values
    :param n: number of outliers
    :param homogeneous: bool, if true returns homogeneous coordinates, otherwise euclidean coordinates, default is true
    :return: np array [(x_1,y_1,1),...,(x_np,y_np,1)] if homogeneous, else [(x_1,y_1),...,(x_np,y_np)]
    """
    n_coords = 3 if homogeneous else 2
    if n <= 0:
        return np.zeros((n, n_coords))
    elif x_range[0] >= x_range[1]:
        raise ValueError("invalid interval for x_range")
    elif y_range[0] >= y_range[1]:
        raise ValueError("invalid interval for y_range")

    outliers_points = []
    for _ in range(n):
        x = random.uniform(*x_range)
        y = random.uniform(*y_range)
        if homogeneous:
            outliers_points.append((x, y, 1))
        else:
            outliers_points.append((x, y))

    return np.array(outliers_points, dtype=float)


# ---------------------------------- CIRCLES -----------------------------------


def circle(
    radius: float,
    center: Tuple[float, float],
    n: int,
    noise_perc: float = 0.0,
    homogeneous: bool = False,
) -> np.ndarray:
    """
    generates circle points given radius and center.

    :param radius: radius
    :param center: center (x,y)
    :param n: number of points
    :param noise_perc: gaussian noise standard deviation
    :param homogeneous: bool, if True returns homogeneous coordinates, otherwise euclidean coordinates, default is False
    :return: np array [(x_1,y_1,1),...,(x_np,y_np,1)]
    """

    # setup
    n_coords = 3 if homogeneous else 2

    points = np.ones((n, n_coords))

    # generate circle's points
    for i in range(n):
        alpha = 2 * math.pi * random.random()
        x = center[0] + radius * math.cos(alpha) + np.random.normal(0, noise_perc)
        y = center[1] + radius * math.sin(alpha) + np.random.normal(0, noise_perc)
        points[i, 0] = x
        points[i, 1] = y

    return points


def circles_sample(
        nm: int,
        n: int,
        noise_perc: float = 0.0,
        outliers_perc: float = 0.0,
        homogeneous: bool = False,
        radius: Union[float, Tuple[float, float]] = None
) -> np.ndarray:
    """

    :param nm: number of models
    :param n: number of points in circle
    :param noise_perc:
    :param outliers_perc:
    :param homogeneous:
    :param radius
    :return:
    """

    # setup
    n_points = int(n * (1 - outliers_perc))
    n_outliers = n - n_points
    n_coords = 3 if homogeneous else 2
    num_inl_per_model = compute_num_inliers_per_model(tot_num_inliers=n_points, num_of_models=nm)
    opt_circle = OPTS["circles"]
    if radius is None:
        min_r, max_r = opt_circle["radius_r"]
    elif type(radius) == float:
        min_r, max_r = radius, radius
    # TODO: FIX dimensionality check
    else:  
        min_r, max_r = radius

    min_x_center, max_x_center = opt_circle["x_center_r"]
    min_center_y, max_center_y = opt_circle["y_center_r"]

    points = np.ones((n_points, n_coords + 1))
    i_tot = 0
    for i_model, npic in zip(range(nm), num_inl_per_model):
        # npic: number of points in circle
        radius = random.uniform(min_r, max_r)
        x_center = random.uniform(min_x_center, max_x_center)
        y_center = random.uniform(min_center_y, max_center_y)
        center = x_center, y_center

        c = circle(
            radius=radius,
            center=center,
            n=npic,
            noise_perc=noise_perc,
            homogeneous=homogeneous
        )

        points[i_tot: i_tot + npic, 0:n_coords] = c
        points[i_tot: i_tot + npic, -1] = i_model + 1
        i_tot += npic

    # generate outliers
    x_range = OPTS["outliers"]["x_r"]
    y_range = OPTS["outliers"]["y_r"]
    outliers_points = outliers(
        x_range=x_range, y_range=y_range, n=n_outliers, homogeneous=homogeneous
    )
    outlier_points_labelled = np.zeros((n_outliers, n_coords+1))
    outlier_points_labelled[:, 0:n_coords] = outliers_points

    points = np.concatenate((points, outlier_points_labelled))
    np.random.shuffle(points)
    return points


def circles_dataset(
    nm: int,
    ns: int,
    n: int,
    noise_perc: float = 0.0,
    outliers_perc: float = 0.0,
    homogeneous: bool = False,
    radius: Union[float, Tuple[float, float], None] = None
) -> np.array:
    """

    :param nm: number of models in each sample
    :param ns: number of samples
    :param n: number of points
    :param noise_perc: gaussian noise standard deviation
    :param outliers_perc: percentage of outliers out of n data points
    :param homogeneous: bool, if True returns homogeneous coordinates, otherwise euclidean coordinates, default is False
    :param radius:
    :return: np.array, shape (ns, n, num coordinates)
    """

    dim = 4 if homogeneous else 3
    samples = np.zeros((ns, n, dim))

    for i in range(ns):

        samples[i] = circles_sample(
            nm=nm,
            n=n,
            noise_perc=noise_perc,
            outliers_perc=outliers_perc,
            homogeneous=homogeneous,
            radius=radius
        )
    return samples


# ---------------------------------- ELLIPSIS -----------------------------------


def ellipse(
    semi_x_axis: float,
    semi_y_axis: float,
    center: Tuple[float, float],
    n: int,
    noise_perc: float = 0.0,
    outliers_perc: float = 0.0,
    homogeneous: bool = True,
    outliers_bounded: bool = True,
) -> np.ndarray:
    """
    samples point from an ellipse.
    Equation is (x-x_0)^2/a^2 + (y-y_0)^2/b^2=1.

    :param semi_x_axis: length of the semi-axis on the abscissa, commonly colled 'a'
    :param semi_y_axis: length of the semi-axis on the ordinate, commonly called 'b'
    :param center: center of the ellipse
    :param n: number of points to be sampled
    :param noise_perc: gaussian noise standard deviation, default is 0.0
    :param outliers_perc:
    :param homogeneous: bool, if true returns homogeneous coordinates, otherwise euclidean coordinates, default is true
    :param outliers_bounded: bool,
    :return: np.array, shape = (n, n_coords)
    """

    # setup
    a = semi_x_axis
    b = semi_y_axis
    x_0, y_0 = center
    x_sampling_interval = (x_0 - a, x_0 + a)
    n_coords = 3 if homogeneous else 2
    n_points = int(n * (1 - outliers_perc))
    n_outliers = n - n_points
    points = np.ones((n_points, n_coords))

    # generate ellipse's points
    for i in range(n_points):
        x = random.uniform(*x_sampling_interval)
        # equation for sampling y: y = sqrt((1-((x-x_0)**2)/a**2)*b**2) + y_0
        y = math.sqrt((1 - ((x - x_0) ** 2 / a**2)) * b**2)
        y = (
            y if random.random() < 0.5 else -y
        )  # choose positive or negative solution of the square root
        y = y + y_0 + np.random.normal(0, noise_perc)
        x = x + np.random.normal(0, noise_perc)
        points[i, 0] = x
        points[i, 1] = y

    # generate outliers
    x_range = (
        (center[0] - a, center[0] + a)
        if outliers_bounded
        else OPTS["outliers"]["x_r"]
    )
    y_range = (
        (center[1] - b, center[1] + b)
        if outliers_bounded
        else OPTS["outliers"]["y_r"]
    )
    outliers_points = outliers(
        x_range=x_range, y_range=y_range, n=n_outliers, homogeneous=homogeneous
    )

    # inliers
    r, c = points.shape
    inliers = np.ones((r, c + 1))
    inliers[:, 0:n_coords] = points

    # outliers
    r, c = outliers_points.shape
    outliers_ = np.zeros((r, c+1))
    outliers_[:, 0: n_coords] = outliers_points

    points = np.concatenate((inliers, outliers_))
    np.random.shuffle(points)

    return points


def ellipsis(
        ns: int,
        n: int,
        noise_perc: float = 0.0,
        outliers_perc: float = 0.0,
        homogeneous: bool = True,
        outliers_bounded: bool = True
) -> np.ndarray:
    """
    Return a numpy array of ns ellipsis

    :param ns: total number of samples
    :param n: number of points to be sampled
    :param noise_perc: gaussian noise standard deviation, default is 0.0
    :param outliers_perc:
    :param homogeneous: bool, if true returns homogeneous coordinates, otherwise euclidean coordinates, default is true
    :param outliers_bounded: bool,
    :return: np.array, shape = (ns, n, 3) if homogeneous = False; (ns, n, 4) if homogeneous = True.
    """
    dim = 3 if not homogeneous else 4
    samples = np.zeros((ns, n, dim))
    opts_ellipse = OPTS["ellipses"]
    min_radius, max_radius = opts_ellipse["radius_r"]
    semi_x_axis = random.uniform(min_radius, max_radius)
    semi_y_axis = random.uniform(min_radius, max_radius)
    min_x_center, max_x_center = opts_ellipse["x_center_r"]
    x_center = random.uniform(min_x_center, max_x_center)
    min_y_center, max_y_center = opts_ellipse["y_center_r"]
    y_center = random.uniform(min_y_center, max_y_center)
    center = x_center, y_center

    for i in range(ns):
        samples[i] = ellipse(
            semi_x_axis=semi_x_axis,
            semi_y_axis=semi_y_axis,
            center=center,
            n=n,
            noise_perc=noise_perc,
            outliers_perc=outliers_perc,
            homogeneous=homogeneous,
            outliers_bounded=outliers_bounded
        )
    return samples


def hyperbola(
    semi_x_axis: float,
    semi_y_axis: float,
    center: Tuple[float, float],
    n: int,
    noise_perc: float = 0.0,
    outliers_perc: float = 0.0,
    homogeneous: bool = True,
    separate: bool = True,
    outliers_bounded: bool = True,
) -> np.ndarray:
    """
    samples point from an ellipse
    :param semi_x_axis: length of the semi-axis on the abscissa, commonly colled 'a'
    :param semi_y_axis: length of the semi-axis on the ordinate, commonly called 'b'
    :param center: center of the ellipse
    :param n: number of points to be sampled
    :param noise_perc: gaussian noise standard deviation, default is 0.0
    :param outliers_perc:
    :param homogeneous: bool, if true returns homogeneous coordinates, otherwise euclidean coordinates, default is true
    :param separate: bool,
    :param outliers_bounded: bool,
    :return: np.array, shape = (n, n_coords)
    """

    # setup
    a = semi_x_axis
    b = semi_y_axis
    (x_0, y_0) = center
    points = []
    n_points = int(n * (1 - outliers_perc))
    n_outliers = n - n_points

    # generate hyperbola's points
    for _ in range(n_points):
        # sample points on the x-axis in such a way they belong in the space of one of the two compontents
        x = random.uniform(x_0 + a, x_0 + 2 * math.sqrt(a**2 + b**2))
        x = x if random.random() < 0.5 else -x
        # equation for sampling y:  y = sqrt((((x-x_0)**2 / a**2)  -  1) * b**2) + y_0
        sqrtarg = (((x - x_0) ** 2 / a**2) - 1) * b**2
        y = math.sqrt(sqrtarg)
        y = (
            y if random.random() < 0.5 else -y
        )  # choose positive or negative solution of the square root
        y = y + y_0 + np.random.normal(0, noise_perc)
        x = x + np.random.normal(0, noise_perc)
        if homogeneous:
            points.append((x, y, 1))
        else:
            points.append((x, y))
    points = np.array(points)

    # generate outliers
    if n_outliers > 0:
        x_outliers_range = (min(points[:, 0]), max(points[:, 0]))
        y_outliers_range = (min(points[:, 1]), max(points[:, 1]))
        x_range = x_outliers_range if outliers_bounded else OPTS["outliers"]["x_r"]
        y_range = y_outliers_range if outliers_bounded else OPTS["outliers"]["y_r"]
        outliers_points = outliers(
            x_range=x_range, y_range=y_range, n=n_outliers, homogeneous=homogeneous
        )
        # process data for output
        if not separate:
            output = np.concatenate((points, outliers_points))
            np.random.shuffle(output)
        else:
            output = points, outliers_points
        return output

    return points


def parabola(
    a: float,
    b: float,
    c: float,
    n: int,
    theta: float = 0,
    noise_perc: float = 0.0,
    outliers_perc: float = 0.0,
    homogeneous: bool = True,
    separate: bool = True,
    outliers_bounded: bool = True,
) -> np.ndarray:
    """
    samples points from a parabola through equation y = ax^2 + bx + c. You can rotate the parabola specifying
    the rotation angle theta ( bounded between 0 and 2*pi).

    :param a:
    :param b:
    :param c:
    :param theta:
    :param n: number of points to be sampled
    :param noise_perc: gaussian noise standard deviation, default is 0.0
    :param outliers_perc:
    :param homogeneous: bool, if true returns homogeneous coordinates, otherwise euclidean coordinates, default is true
    :param separate: bool,
    :param outliers_bounded: bool,
    :return: np.array, shape = (n, n_coords)
    """

    # setup
    points = []
    n_points = int(n * (1 - outliers_perc))
    n_outliers = n - n_points
    x_vertex = -b / 2 * a
    amplitude = 2 / (0.1 + a)
    x_sampling_interval = (x_vertex - amplitude, x_vertex + amplitude)

    # generate parabola's points
    for _ in range(n_points):
        x = random.uniform(*x_sampling_interval)
        y = a * x**2 + b * x + c
        xp = x * math.cos(theta) - y * math.sin(theta) + np.random.normal(0, noise_perc)
        yp = y * math.cos(theta) + x * math.sin(theta) + np.random.normal(0, noise_perc)

        if homogeneous:
            points.append((xp, yp, 1))
        else:
            points.append((xp, yp))
    points = np.array(points)

    # generate outliers
    if n_outliers > 0:
        x_outliers_range = (
            min(point[0] for point in points),
            max(point[0] for point in points),
        )
        y_outliers_range = (
            min(point[1] for point in points),
            max(point[1] for point in points),
        )
        x_range = x_outliers_range if outliers_bounded else OPTS["outliers"]["x_r"]
        y_range = y_outliers_range if outliers_bounded else OPTS["outliers"]["y_r"]
        outliers_points = outliers(
            x_range=x_range, y_range=y_range, n=n_outliers, homogeneous=homogeneous
        )

        # process data for output
        if not separate:
            output = np.concatenate((points, outliers_points))
            np.random.shuffle(output)
        else:
            output = points, outliers_points
        return output
    return points


def line(
    point_1: Tuple[float, float],
    point_2: Tuple[float, float],
    n: int,
    noise_perc: float = 0.0,
    outliers_perc: float = 0.0,
    homogeneous: bool = True,
    separate: bool = True,
    outliers_bounded: bool = True,
):
    """
    samples point from a line obtained by connecting two points
    :param point_1: (float, float)
    :param point_2: (float, float)
    :param n: number of points to be drawn
    :param noise_perc: gaussian noise standard deviation, default is 0.0
    :param outliers_perc:
    :param homogeneous: bool, if true returns homogeneous coordinates, otherwise euclidean coordinates, default is true
    :param separate: bool,
    :param outliers_bounded: bool,
    :return: np.array, shape = (n, n_coords)
    """

    # setup
    x1, y1 = point_1
    x2, y2 = point_2
    m = (y1 - y2) / (x1 - x2)
    q = y1 - m * x1
    points = []
    n_points = int(n * (1 - outliers_perc))
    n_outliers = n - n_points

    # generate line's points
    for _ in range(n_points):
        x = random.uniform(point_1[0], point_2[0])
        y = m * x + q + np.random.normal(0, noise_perc)
        if homogeneous:
            points.append((x, y, 1))
        else:
            points.append((x, y))
    points = np.array(points)

    # generate outliers
    if n_outliers > 0:
        x_range = (
            (min(points[:, 0]), max(points[:, 0]))
            if outliers_bounded
            else OPTS["outliers"]["x_r"]
        )
        y_range = (
            (min(points[:, 1]), max(points[:, 1]))
            if outliers_bounded
            else OPTS["outliers"]["y_r"]
        )
        outliers_points = outliers(
            x_range=x_range, y_range=y_range, n=n_outliers, homogeneous=homogeneous
        )

        # process data for output
        if not separate:
            output = np.concatenate((points, outliers_points))
            np.random.shuffle(output)
        else:
            output = points, outliers_points
        return output
    return points


def homography_v1(
    matrix: Union[List, np.array],
    n: int,
    src_range_x: Tuple[float, float] = (-5, 5),
    src_range_y: Tuple[float, float] = (-5, 5),
    src_range_z: Tuple[float, float] = (-5, 5),
    outliers_perc: float = 0.0,
    noise_perc: float = 0.0,
    homogeneous: bool = True,
    separate: bool = True,
):
    """
    generates points belonging to homography specified as input
    :param matrix: list 3x3 or np.array with shape (3,3)
    :param n: number of points to be drawn
    :param src_range_x:
    :param src_range_y:
    :param src_range_z:
    :param outliers_perc:
    :param noise_perc:
    :param homogeneous:
    :param separate:
    :return:
    """

    # setup
    n_points = int(n * (1 - outliers_perc))
    n_outliers = n - n_points

    # generate homography points
    x1s = np.random.uniform(
        low=(src_range_x[0], src_range_y[0], src_range_z[0]),
        high=(src_range_x[1], src_range_y[1], src_range_z[1]),
        size=(n_points, 2),
    )  # todo why 2??
    x1s = x1s / x1s[:, 2]
    x2s = np.matmul(x1s, matrix)
    x2s = x2s / x2s[:, 2]
    x1s[:, 0:2] = x1s[:, 0:2] + np.random.normal(
        loc=0, scale=noise_perc, size=(n_points, 2)
    )  # add noise
    x2s[:, 0:2] = x2s[:, 0:2] + np.random.normal(
        loc=0, scale=noise_perc, size=(n_points, 2)
    )  # add noise
    if homogeneous:
        points = np.concatenate((x1s, x2s))
    else:
        points = np.concatenate((x1s[0:2], x2s[0:2]))

    # generate outliers
    if n_outliers > 0:
        outliers_points = np.random.uniform(
            low=(src_range_x[0], src_range_y[0], src_range_z[0]),
            high=(src_range_x[1], src_range_y[1], src_range_z[1]),
            size=(n_outliers, 2),
        )
        outliers_points = outliers_points / outliers_points[:, 2]

        # process data for output
        if separate:
            output = np.concatenate((points, outliers_points))
            np.random.shuffle(output)
        else:
            output = points, outliers_points
        return output
    return points


def homography(
    matrix: Union[List, np.array],
    n: int,
    src_range_x: Tuple[float, float] = (-5, 5),
    src_range_y: Tuple[float, float] = (-5, 5),
    src_range_z: Tuple[float, float] = (-5, 5),
    outliers_perc: float = 0.0,
    noise_perc: float = 0.0,
    homogeneous: bool = True,
    separate: bool = True,
):
    """
    generates points belonging to homography specified as input
    :param matrix: list 3x3 or np.array with shape (3,3)
    :param n: number of points to be drawn
    :param src_range_x:
    :param src_range_y:
    :param src_range_z:
    :param outliers_perc:
    :param noise_perc:
    :param homogeneous:
    :param separate:
    :return:
    """

    # setup
    n_points = int(n * (1 - outliers_perc))
    n_outliers = n - n_points

    random_point_cloud = np.random.uniform(
        low=(src_range_x[0], src_range_y[0], src_range_z[0]),
        high=(src_range_x[1], src_range_y[1], src_range_z[1]),
        size=(3, n_points),
    )

    # rotate point cloud
    angles = 0.8 * (0.5 - np.random.rand(1, 3))
    R = Rotation.from_euler("zyx", angles, degrees=False).as_matrix()[0]
    t = np.zeros((3, 1))

    # generate camera projection
    x1p_org, x2p_org, K = simCamProj("uncalibrated", random_point_cloud, R, t)


def homography_from_pointcloud(X: np.ndarray, matrix, rotation_threshold: float = 0.8):
    """
    starting from a 3D point clouds computes a pair of 2D projections explainable by a homography

    apply random rotation R1 to X,
    then generates 2 2D-projections of X:
    - the first  one is a projection of X rotated by Rglobal,
    - the second one is a projection obtained by further applying R and t

    :param X: 3D point cloud, np.ndarray, (3, npts)
    :return:
        x1ph: 1st projection in homogeneous coordinates, np.ndarray, (3, nps)
        x2ph: 2nd projection in homogeneous coordinates, np.ndarray, (3, nps)
        TN1:  de-normalization matrix for 1st projection supposedly, np.ndarray, (3,3)
        TN2:  de-normalization matrix for 2nd projection supposedly, np.ndarray, (3,3)
        Forg: identity, np.ndarray, (3,3)
        Fn:   identity, np.ndarray, (3,3)
        E:    identity, np.ndarray, (3,3)
        K,    intrinsic matrix, np.ndarray, (3,3)
        R,    motion matrix from view 1 to view 2, np.ndarray, (3,3)
        t,    translation vector from view 1 to view 2, np.ndarray, (3,3)
        H,    homography matrix, np.ndarray, (3,3)
        Hn,   homography matrix to be applied to normalized data, np.ndarray, (3,3)
    """
    motion = "rotation"
    camproj = "uncalibrated"
    npts = X.shape[-1]

    # randomly rotate X
    angles = 0.8 * (0.5 - np.random.rand(1, 3))
    Rglobal = Rotation.from_euler("zyx", angles, degrees=False)
    Rglobal = Rglobal.as_matrix()[0]
    X = np.matmul(Rglobal, X)

    # put in front of camera 1
    X = X + np.repeat(np.transpose(np.array([[0, 0, 2]])), npts, axis=1)
    # mix homographies
    # skipped

    # generate motion
    R, t = simCamTransform(motion, X, rotation_threshold=rotation_threshold)

    # generate camera projection
    x1p_org, x2p_org, K = simCamProj(camproj, X, R, t)

    x1p, TN1 = normalize(x1p_org, 1)
    x2p, TN2 = normalize(x2p_org, 1)

    x1ph = np.ones((3, npts))
    x1ph[0:2] = x1p
    x2ph = np.ones((3, npts))
    x2ph[0:2] = x2p

    # add noise
    noise = 10 / 512
    x1ph[0:2, :] = x1ph[0:2, :] + (2 * np.random.rand(2, npts) - 1) * noise
    x2ph[0:2, :] = x2ph[0:2, :] + (2 * np.random.rand(2, npts) - 1) * noise

    # compute homography from rotation
    H = np.matmul(K, np.matmul(R, np.linalg.inv(K)))
    Hn = np.matmul(TN2, np.matmul(np.linalg.inv(H), np.linalg.inv(TN1)))

    E = np.eye(3)
    Forg = np.eye(3)
    Fn = np.eye(3)

    return x1ph, x2ph, TN1, TN2, Forg, Fn, E, K, R, t, H, Hn



import pandas as pd
from dash import Dash, dcc, html, Input, Output
import plotly.express as px

from sydraw import synth

app = Dash(__name__)


app.layout = html.Div([
    dcc.Graph(id="scatter-plot",
              style={'width': '90vh',
                     'height': '90vh',
                     'float': 'left',
                     'show_scale': 'false'}),

    # Number of models
    html.Div([
        html.P("Number of models"),
        dcc.Slider(
            id='nm',
            min=1, max=5, step=1,
            marks={0: '0', 1: '1', 2: '2', 3: '3', 4: '4', 5: '5'},
            value=2,
        ),

        # Number of points
        html.P("Number of points"),
        dcc.Slider(
                id='n',
                min=100, max=1000, step=100,
                marks={100: '100',
                       200: '200',
                       300: '300',
                       400: '400',
                       500: '500',
                       600: '600',
                       700: '700',
                       800: '800',
                       900: '900',
                       1000: '1000'},
                value=500
            ),

        # Noise
        html.P("Noise percentage"),
        dcc.Slider(
                id='noise_perc',
                min=0, max=0.1, step=0.01,
                marks={0: '0.0',
                       0.01: '0.01',
                       0.02: '0.02',
                       0.03: '0.03',
                       0.04: '0.04',
                       0.05: '0.05',
                       0.06: '0.06',
                       0.07: '0.07',
                       0.08: '0.08',
                       0.09: '0.09',
                       0.1: '0.10'},
                value=0.01
            ),

        # Outliers
        html.P("Outliers percentage"),
        dcc.Slider(
                id='outliers_perc',
                min=0.1, max=1, step=0.10,
                marks={0: '0.0',
                       0.1: '0.1',
                       0.2: '0.2',
                       0.3: '0.3',
                       0.4: '0.4',
                       0.5: '0.5',
                       0.6: '0.6',
                       0.7: '0.7',
                       0.8: '0.8',
                       0.9: '0.9',
                       1: '1.0'},
                value=0.1
            ),
        html.P("Radii range"),
                dcc.RangeSlider(
                        id='radius',
                        min=0.1, max=1.0, step=0.10,
                        marks={0: '0.0',
                               0.1: '0.1',
                               0.2: '0.2',
                               0.3: '0.3',
                               0.4: '0.4',
                               0.5: '0.5',
                               0.6: '0.6',
                               0.7: '0.7',
                               0.8: '0.8',
                               0.9: '0.9',
                               1: '1.0'},
                        value=[0.3, 0.5]
                    )
        ],
    style={'width': '90vh',
           'height': '90vh',
           'float': 'right',
           'padding-top': '10%'})
])


@app.callback(
    Output("scatter-plot", "figure"),
    Input("nm", "value"),
    Input("n", "value"),
    Input("noise_perc", "value"),
    Input("outliers_perc", "value"),
    Input("radius", "value"))
def update_bar_chart(nm, n, noise_perc, outliers_perc, radius):
    sample = synth.circles_sample(
        nm=nm,
        n=n,
        noise_perc=noise_perc,
        outliers_perc=outliers_perc,
        radius=radius,
        homogeneous=False
    )
    df = pd.DataFrame(sample, columns=["x", "y", "label"])
    fig = px.scatter(
        df,
        x="x",
        y="y",
        color="label",
        range_x=(-2.5, 2.5),
        range_y=(-2.5, 2.5),
    template='simple_white')
    fig.update_coloraxes(showscale=False)
    fig.update_traces(marker={'size': 3})

    return fig


if __name__ == "__main__":
    app.run_server(debug=True)

